@@ -174,7 +174,11 @@ xkl_set_log_appender(XklLogAppender func)
 gint
 xkl_engine_start_listen(XklEngine * engine, guint what)
 {
-	xkl_engine_priv(engine, listener_type) = what;
+	int i;
+	guchar *cntr = xkl_engine_priv(engine, listener_type_counters);
+	for (i = 0; i < XKLL_NUMBER_OF_LISTEN_MODES; i++, cntr++)
+		if (what & (1 << i))
+			(*cntr)++;
 
 	if (!
 	    (xkl_engine_priv(engine, features) &
@@ -182,7 +186,7 @@ xkl_engine_start_listen(XklEngine * engine, guint what)
 && (what & XKLL_MANAGE_LAYOUTS))
 		xkl_debug(0,
 			  "The backend does not require manual layout management - "
-			  "but it is provided by the application");
+			  "but it is provided by the application\n");
 
 	xkl_engine_resume_listen(engine);
 	xkl_engine_load_window_tree(engine);
@@ -191,9 +195,22 @@ xkl_engine_start_listen(XklEngine * engine, guint what)
 }
 
 gint
-xkl_engine_stop_listen(XklEngine * engine)
+xkl_engine_stop_listen(XklEngine * engine, guint what)
 {
-	xkl_engine_pause_listen(engine);
+	int i;
+	gboolean no_flags_remains = TRUE;
+	guchar *cntr = xkl_engine_priv(engine, listener_type_counters);
+	for (i = 0; i < XKLL_NUMBER_OF_LISTEN_MODES; i++, cntr++) {
+		int mask = 1 << i;
+		if (what & mask)
+			(*cntr)--;
+
+		if (*cntr)
+			no_flags_remains = FALSE;
+	}
+	if (no_flags_remains)
+		xkl_engine_pause_listen(engine);
+
 	return 0;
 }
 
@@ -341,8 +358,7 @@ xkl_engine_load_window_tree(XklEngine * engine)
 	int revert;
 	gboolean retval = TRUE, have_toplevel_win;
 
-	if (xkl_engine_priv(engine, listener_type) &
-	    XKLL_MANAGE_WINDOW_STATES)
+	if (xkl_engine_is_listening_for(engine, XKLL_MANAGE_WINDOW_STATES))
 		retval =
 		    xkl_engine_load_subtree(engine,
 					    xkl_engine_priv(engine,
@@ -560,6 +576,7 @@ xkl_engine_lock_group(XklEngine * engine, int group)
 gint
 xkl_engine_pause_listen(XklEngine * engine)
 {
+	xkl_debug(150, "Pause listening\n");
 	xkl_engine_ensure_vtable_inited(engine);
 	return xkl_engine_vcall(engine, pause_listen) (engine);
 }
@@ -568,8 +585,17 @@ gint
 xkl_engine_resume_listen(XklEngine * engine)
 {
 	xkl_engine_ensure_vtable_inited(engine);
-	xkl_debug(150, "listenerType: %x\n",
-		  xkl_engine_priv(engine, listener_type));
+	guchar *listener_type_counters =
+	    xkl_engine_priv(engine, listener_type_counters);
+	xkl_debug(150, "Resume listening, listenerType: (%s%s%s)\n",
+		  (listener_type_counters
+		   [XKLL_MANAGE_WINDOW_STATES_OFFSET]) ?
+		  "XKLL_MANAGE_WINDOW_STATES " : "",
+		  (listener_type_counters
+		   [XKLL_TRACK_KEYBOARD_STATE_OFFSET]) ?
+		  "XKLL_TRACK_KEYBOARD_STATE " : "",
+		  (listener_type_counters[XKLL_MANAGE_LAYOUTS_OFFSET]) ?
+		  "XKLL_MANAGE_LAYOUTS " : "");
 	if (xkl_engine_vcall(engine, resume_listen) (engine))
 		return 1;
 
@@ -624,8 +650,7 @@ xkl_engine_constructor(GType type,
 
 	engine = XKL_ENGINE(obj);
 
-	display =
-	    (Display *)
+	display = (Display *)
 	    g_value_peek_pointer(construct_properties[0].value);
 
 	xkl_engine_priv(engine, display) = display;
@@ -671,7 +696,8 @@ xkl_engine_constructor(GType type,
 			  xkl_engine_get_backend_name(engine));
 	} else {
 		xkl_debug(0, "All backends failed, last result: %d\n", rv);
-		XSetErrorHandler(xkl_engine_priv(engine, default_error_handler));
+		XSetErrorHandler(xkl_engine_priv
+				 (engine, default_error_handler));
 		xkl_engine_priv(engine, display) = NULL;
 		g_object_unref(G_OBJECT(engine));
 		return NULL;
