@@ -35,12 +35,17 @@ static xmlXPathCompExprPtr models_xpath;
 static xmlXPathCompExprPtr layouts_xpath;
 static xmlXPathCompExprPtr option_groups_xpath;
 
+static GRegex **xml_encode_regexen = NULL;
+static GRegex **xml_decode_regexen = NULL;
+static const char *xml_decode_regexen_str[] = { "&lt;", "&gt;", "&amp;" };
+static const char *xml_encode_regexen_str[] = { "<", ">", "&" };
+
 /* gettext domain for translations */
 #define XKB_DOMAIN "xkeyboard-config"
 
 enum {
 	PROP_0,
-	PROP_ENGINE,
+	PROP_ENGINE
 };
 
 static gboolean
@@ -71,7 +76,7 @@ xkl_xml_find_config_item_child(xmlNodePtr iptr, xmlNodePtr * ptr)
 }
 
 static xmlNodePtr
-xkl_find_nonlocalized_element(xmlNodePtr ptr, const gchar * tag_name)
+xkl_find_element(xmlNodePtr ptr, const gchar * tag_name)
 {
 	xmlNodePtr found_element = NULL;
 
@@ -79,14 +84,9 @@ xkl_find_nonlocalized_element(xmlNodePtr ptr, const gchar * tag_name)
 	while (ptr != NULL) {
 		char *node_name = (char *) ptr->name;
 		if (ptr->type != XML_TEXT_NODE) {
-			xmlChar *lang = xmlNodeGetLang(ptr);
-
-			if (lang == NULL) {	/* No language specified */
-				if (!g_ascii_strcasecmp
-				    (node_name, tag_name)) {
-					found_element = ptr;
-					break;
-				}
+			if (!g_ascii_strcasecmp(node_name, tag_name)) {
+				found_element = ptr;
+				break;
 			}
 		}
 		ptr = ptr->next;
@@ -101,7 +101,7 @@ xkl_item_populate_optional_array(XklConfigItem * item, xmlNodePtr ptr,
 				 const gchar property_name[])
 {
 	xmlNodePtr top_list_element =
-	    xkl_find_nonlocalized_element(ptr, list_tag), element_ptr;
+	    xkl_find_element(ptr, list_tag), element_ptr;
 	gint n_elements, idx;
 	gchar **elements = NULL;
 
@@ -114,7 +114,7 @@ xkl_item_populate_optional_array(XklConfigItem * item, xmlNodePtr ptr,
 	element_ptr = top_list_element->children;
 	while (NULL !=
 	       (element_ptr =
-		xkl_find_nonlocalized_element(element_ptr, element_tag))) {
+		xkl_find_element(element_ptr, element_tag))) {
 		n_elements++;
 		element_ptr = element_ptr->next;
 	}
@@ -127,8 +127,7 @@ xkl_item_populate_optional_array(XklConfigItem * item, xmlNodePtr ptr,
 	element_ptr = top_list_element->children;
 	for (idx = 0;
 	     NULL != (element_ptr =
-		      xkl_find_nonlocalized_element
-		      (element_ptr, element_tag));
+		      xkl_find_element(element_ptr, element_tag));
 	     element_ptr = element_ptr->next, idx++) {
 		elements[idx] =
 		    g_strdup((const char *) element_ptr->children->
@@ -141,15 +140,20 @@ xkl_item_populate_optional_array(XklConfigItem * item, xmlNodePtr ptr,
 	return TRUE;
 }
 
+#include "libxml/parserInternals.h"
+
 gboolean
-xkl_read_config_item(XklConfigRegistry * config, xmlNodePtr iptr,
-		     XklConfigItem * item)
+xkl_read_config_item(XklConfigRegistry * config, gint doc_index,
+		     xmlNodePtr iptr, XklConfigItem * item)
 {
 	xmlNodePtr name_element, ptr;
 	xmlNodePtr desc_element = NULL, short_desc_element =
 	    NULL, vendor_element = NULL;
 
-	gchar *vendor = NULL;
+	gchar *vendor = NULL, *translated = NULL, *escaped =
+	    NULL, *unescaped = NULL;
+
+	gint i;
 
 	*item->name = 0;
 	*item->short_description = 0;
@@ -162,6 +166,10 @@ xkl_read_config_item(XklConfigRegistry * config, xmlNodePtr iptr,
 	if (!xkl_xml_find_config_item_child(iptr, &ptr))
 		return FALSE;
 
+	if (doc_index > 0)
+		g_object_set_data(G_OBJECT(item), XCI_PROP_EXTRA_ITEM,
+				  GINT_TO_POINTER(TRUE));
+
 	ptr = ptr->children;
 
 	if (ptr->type == XML_TEXT_NODE)
@@ -169,11 +177,9 @@ xkl_read_config_item(XklConfigRegistry * config, xmlNodePtr iptr,
 	name_element = ptr;
 	ptr = ptr->next;
 
-	short_desc_element =
-	    xkl_find_nonlocalized_element(ptr, XML_TAG_SHORT_DESCR);
-	desc_element = xkl_find_nonlocalized_element(ptr, XML_TAG_DESCR);
-	vendor_element =
-	    xkl_find_nonlocalized_element(ptr, XML_TAG_VENDOR);
+	short_desc_element = xkl_find_element(ptr, XML_TAG_SHORT_DESCR);
+	desc_element = xkl_find_element(ptr, XML_TAG_DESCR);
+	vendor_element = xkl_find_element(ptr, XML_TAG_VENDOR);
 
 	if (name_element != NULL && name_element->children != NULL)
 		strncat(item->name,
@@ -183,18 +189,49 @@ xkl_read_config_item(XklConfigRegistry * config, xmlNodePtr iptr,
 	if (short_desc_element != NULL
 	    && short_desc_element->children != NULL) {
 		strncat(item->short_description,
-			dgettext(XKB_DOMAIN,
-				 (const char *)
+			dgettext(XKB_DOMAIN, (const char *)
 				 short_desc_element->children->content),
 			XKL_MAX_CI_SHORT_DESC_LENGTH - 1);
 	}
 
 	if (desc_element != NULL && desc_element->children != NULL) {
+		/* Convert all xml-related characters to XML form, otherwise dgettext won't find the translation 
+		 * The conversion is not using libxml2, because there are no handy functions in API */
+		translated =
+		    g_strdup((gchar *) desc_element->children->content);
+		for (i =
+		     sizeof(xml_encode_regexen_str) /
+		     sizeof(xml_encode_regexen_str[0]); --i >= 0;) {
+			escaped =
+			    g_regex_replace(xml_encode_regexen[i],
+					    translated, -1, 0,
+					    xml_decode_regexen_str[i], 0,
+					    NULL);
+			g_free(translated);
+			translated = escaped;
+		}
+		escaped = translated;
+
+		/* Do the translation! */
+		translated =
+		    g_strdup(dgettext(XKB_DOMAIN, (const char *) escaped));
+		g_free(escaped);
+
+		/* Convert all XML entities back to normal form */
+		for (i =
+		     sizeof(xml_decode_regexen_str) /
+		     sizeof(xml_decode_regexen_str[0]); --i >= 0;) {
+			unescaped =
+			    g_regex_replace(xml_decode_regexen[i],
+					    translated, -1, 0,
+					    xml_encode_regexen_str[i], 0,
+					    NULL);
+			g_free(translated);
+			translated = unescaped;
+		}
 		strncat(item->description,
-			dgettext(XKB_DOMAIN,
-				 (const char *) desc_element->
-				 children->content),
-			XKL_MAX_CI_DESC_LENGTH - 1);
+			translated, XKL_MAX_CI_DESC_LENGTH - 1);
+		g_free(translated);
 	}
 
 	if (vendor_element != NULL && vendor_element->children != NULL) {
@@ -217,7 +254,8 @@ xkl_read_config_item(XklConfigRegistry * config, xmlNodePtr iptr,
 
 static void
 xkl_config_registry_foreach_in_nodeset(XklConfigRegistry * config,
-				       xmlNodeSetPtr nodes,
+				       GSList ** processed_ids,
+				       gint doc_index, xmlNodeSetPtr nodes,
 				       ConfigItemProcessFunc func,
 				       gpointer data)
 {
@@ -226,8 +264,19 @@ xkl_config_registry_foreach_in_nodeset(XklConfigRegistry * config,
 		xmlNodePtr *pnode = nodes->nodeTab;
 		XklConfigItem *ci = xkl_config_item_new();
 		for (i = nodes->nodeNr; --i >= 0;) {
-			if (xkl_read_config_item(config, *pnode, ci))
-				func(config, ci, data);
+			if (xkl_read_config_item
+			    (config, doc_index, *pnode, ci)) {
+				if (g_slist_find_custom
+				    (*processed_ids, ci->name,
+				     (GCompareFunc) g_ascii_strcasecmp) ==
+				    NULL) {
+					func(config, ci, data);
+					*processed_ids =
+					    g_slist_append(*processed_ids,
+							   g_strdup
+							   (ci->name));
+				}
+			}
 
 			pnode++;
 		}
@@ -243,18 +292,30 @@ xkl_config_registry_foreach_in_xpath(XklConfigRegistry * config,
 				     gpointer data)
 {
 	xmlXPathObjectPtr xpath_obj;
+	gint di;
+	GSList *processed_ids = NULL;
 
 	if (!xkl_config_registry_is_initialized(config))
 		return;
-	xpath_obj = xmlXPathCompiledEval(xpath_comp_expr,
-					 xkl_config_registry_priv
-					 (config, xpath_context));
-	if (xpath_obj != NULL) {
+
+	for (di = 0; di < XKL_NUMBER_OF_REGISTRY_DOCS; di++) {
+		xmlXPathContextPtr xmlctxt =
+		    xkl_config_registry_priv(config, xpath_contexts[di]);
+		if (xmlctxt == NULL)
+			continue;
+
+		xpath_obj = xmlXPathCompiledEval(xpath_comp_expr, xmlctxt);
+		if (xpath_obj == NULL)
+			continue;
+
 		xkl_config_registry_foreach_in_nodeset(config,
+						       &processed_ids, di,
 						       xpath_obj->nodesetval,
 						       func, data);
 		xmlXPathFreeObject(xpath_obj);
 	}
+	g_slist_foreach(processed_ids, (GFunc) g_free, NULL);
+	g_slist_free(processed_ids);
 }
 
 void
@@ -269,19 +330,33 @@ xkl_config_registry_foreach_in_xpath_with_param(XklConfigRegistry
 {
 	char xpath_expr[1024];
 	xmlXPathObjectPtr xpath_obj;
+	gint di;
+	GSList *processed_ids = NULL;
 
 	if (!xkl_config_registry_is_initialized(config))
 		return;
+
 	g_snprintf(xpath_expr, sizeof xpath_expr, format, value);
-	xpath_obj = xmlXPathEval((unsigned char *) xpath_expr,
-				 xkl_config_registry_priv(config,
-							  xpath_context));
-	if (xpath_obj != NULL) {
+
+	for (di = 0; di < XKL_NUMBER_OF_REGISTRY_DOCS; di++) {
+		xmlXPathContextPtr xmlctxt =
+		    xkl_config_registry_priv(config, xpath_contexts[di]);
+		if (xmlctxt == NULL)
+			continue;
+
+		xpath_obj =
+		    xmlXPathEval((unsigned char *) xpath_expr, xmlctxt);
+		if (xpath_obj == NULL)
+			continue;
+
 		xkl_config_registry_foreach_in_nodeset(config,
+						       &processed_ids, di,
 						       xpath_obj->nodesetval,
 						       func, data);
 		xmlXPathFreeObject(xpath_obj);
 	}
+	g_slist_foreach(processed_ids, (GFunc) g_free, NULL);
+	g_slist_free(processed_ids);
 }
 
 static gboolean
@@ -295,28 +370,38 @@ xkl_config_registry_find_object(XklConfigRegistry * config,
 	xmlNodeSetPtr nodes;
 	gboolean rv = FALSE;
 	gchar xpath_expr[1024];
+	gint di;
 
 	if (!xkl_config_registry_is_initialized(config))
 		return FALSE;
 
 	g_snprintf(xpath_expr, sizeof xpath_expr, format, arg1,
 		   pitem->name);
-	xpath_obj =
-	    xmlXPathEval((unsigned char *) xpath_expr,
-			 xkl_config_registry_priv(config, xpath_context));
-	if (xpath_obj == NULL)
-		return FALSE;
 
-	nodes = xpath_obj->nodesetval;
-	if (nodes != NULL && nodes->nodeTab != NULL && nodes->nodeNr > 0) {
-		rv = xkl_read_config_item(config, nodes->nodeTab[0],
-					  pitem);
-		if (pnode != NULL) {
-			*pnode = *nodes->nodeTab;
+	for (di = 0; di < XKL_NUMBER_OF_REGISTRY_DOCS; di++) {
+		xmlXPathContextPtr xmlctxt =
+		    xkl_config_registry_priv(config, xpath_contexts[di]);
+		if (xmlctxt == NULL)
+			continue;
+
+		xpath_obj =
+		    xmlXPathEval((unsigned char *) xpath_expr, xmlctxt);
+		if (xpath_obj == NULL)
+			continue;
+
+		nodes = xpath_obj->nodesetval;
+		if (nodes != NULL && nodes->nodeTab != NULL
+		    && nodes->nodeNr > 0) {
+			rv = xkl_read_config_item(config, di,
+						  nodes->nodeTab[0],
+						  pitem);
+			if (pnode != NULL) {
+				*pnode = *nodes->nodeTab;
+			}
 		}
-	}
 
-	xmlXPathFreeObject(xpath_obj);
+		xmlXPathFreeObject(xpath_obj);
+	}
 	return rv;
 }
 
@@ -476,9 +561,11 @@ xkl_xml_sax_end_element_ns(void *ctx,
 
 gboolean
 xkl_config_registry_load_from_file(XklConfigRegistry * config,
-				   const gchar * file_name)
+				   const gchar * file_name, gint docidx)
 {
 	xmlParserCtxtPtr ctxt = xmlNewParserCtxt();
+	xmlDocPtr doc;
+
 	xkl_debug(100, "Loading XML registry from file %s\n", file_name);
 
 	/* Filter out all unneeded languages! */
@@ -487,18 +574,20 @@ xkl_config_registry_load_from_file(XklConfigRegistry * config,
 	ctxt->sax->endElementNs = xkl_xml_sax_end_element_ns;
 	ctxt->sax->characters = xkl_xml_sax_characters;
 
-	xkl_config_registry_priv(config, doc) =
+	doc = xkl_config_registry_priv(config, docs[docidx]) =
 	    xmlCtxtReadFile(ctxt, file_name, NULL, XML_PARSE_NOBLANKS);
 	xmlFreeParserCtxt(ctxt);
 
-	if (xkl_config_registry_priv(config, doc) == NULL) {
-		xkl_config_registry_priv(config, xpath_context) = NULL;
+	if (doc == NULL) {
+		xkl_config_registry_priv(config, xpath_contexts[docidx]) =
+		    NULL;
 		xkl_last_error_message =
-		    "Could not parse XKB configuration registry";
+		    "Could not parse primary XKB configuration registry";
 		return FALSE;
 	}
-	xkl_config_registry_priv(config, xpath_context) =
-	    xmlXPathNewContext(xkl_config_registry_priv(config, doc));
+
+	xkl_config_registry_priv(config, xpath_contexts[docidx]) =
+	    xmlXPathNewContext(doc);
 
 	return TRUE;
 }
@@ -506,7 +595,8 @@ xkl_config_registry_load_from_file(XklConfigRegistry * config,
 gboolean
 xkl_config_registry_load_helper(XklConfigRegistry * config,
 				const char default_ruleset[],
-				const char base_dir[])
+				const char base_dir[],
+				gboolean if_extras_needed)
 {
 	struct stat stat_buf;
 	gchar file_name[MAXPATHLEN] = "";
@@ -524,18 +614,43 @@ xkl_config_registry_load_helper(XklConfigRegistry * config,
 		return FALSE;
 	}
 
-	return xkl_config_registry_load_from_file(config, file_name);
+	if (!xkl_config_registry_load_from_file(config, file_name, 0))
+		return FALSE;
+
+	if (!if_extras_needed)
+		return TRUE;
+
+	g_snprintf(file_name, sizeof file_name, "%s/%s.extras.xml",
+		   base_dir, rf);
+
+	/* no extras - ok, no problem */
+	if (stat(file_name, &stat_buf) != 0)
+		return TRUE;
+
+	return xkl_config_registry_load_from_file(config, file_name, 1);
 }
 
 void
 xkl_config_registry_free(XklConfigRegistry * config)
 {
 	if (xkl_config_registry_is_initialized(config)) {
-		xmlXPathFreeContext(xkl_config_registry_priv
-				    (config, xpath_context));
-		xmlFreeDoc(xkl_config_registry_priv(config, doc));
-		xkl_config_registry_priv(config, xpath_context) = NULL;
-		xkl_config_registry_priv(config, doc) = NULL;
+		gint di;
+		for (di = 0; di < XKL_NUMBER_OF_REGISTRY_DOCS; di++) {
+			xmlXPathContextPtr xmlctxt =
+			    xkl_config_registry_priv(config,
+						     xpath_contexts[di]);
+			if (xmlctxt == NULL)
+				continue;
+
+			xmlXPathFreeContext(xmlctxt);
+			xmlFreeDoc(xkl_config_registry_priv
+				   (config, docs[di]));
+			xkl_config_registry_priv(config,
+						 xpath_contexts[di]) =
+			    NULL;
+			xkl_config_registry_priv(config, docs[di]) = NULL;
+		}
+
 	}
 }
 
@@ -579,40 +694,61 @@ xkl_config_registry_foreach_option_group(XklConfigRegistry *
 					 func, gpointer data)
 {
 	xmlXPathObjectPtr xpath_obj;
-	gint i;
+	gint di, j;
+	GSList *processed_ids = NULL;
 
 	if (!xkl_config_registry_is_initialized(config))
 		return;
-	xpath_obj =
-	    xmlXPathCompiledEval(option_groups_xpath,
-				 xkl_config_registry_priv(config,
-							  xpath_context));
-	if (xpath_obj != NULL) {
-		xmlNodeSetPtr nodes = xpath_obj->nodesetval;
-		xmlNodePtr *pnode = nodes->nodeTab;
-		XklConfigItem *ci = xkl_config_item_new();
-		for (i = nodes->nodeNr; --i >= 0;) {
 
-			if (xkl_read_config_item(config, *pnode, ci)) {
-				gboolean allow_multisel = TRUE;
-				xmlChar *sallow_multisel =
-				    xmlGetProp(*pnode,
-					       (unsigned char *)
-					       XCI_PROP_ALLOW_MULTIPLE_SELECTION);
-				if (sallow_multisel != NULL) {
-					allow_multisel =
-					    !g_ascii_strcasecmp
-					    ("true", (char *)
-					     sallow_multisel);
-					xmlFree(sallow_multisel);
-					g_object_set_data(G_OBJECT
-							  (ci),
-							  XCI_PROP_ALLOW_MULTIPLE_SELECTION,
-							  GINT_TO_POINTER
-							  (allow_multisel));
+	for (di = 0; di < XKL_NUMBER_OF_REGISTRY_DOCS; di++) {
+		xmlNodeSetPtr nodes;
+		xmlNodePtr *pnode;
+		XklConfigItem *ci;
+
+		xmlXPathContextPtr xmlctxt =
+		    xkl_config_registry_priv(config, xpath_contexts[di]);
+		if (xmlctxt == NULL)
+			continue;
+
+		xpath_obj =
+		    xmlXPathCompiledEval(option_groups_xpath, xmlctxt);
+		if (xpath_obj == NULL)
+			continue;
+
+		nodes = xpath_obj->nodesetval;
+		pnode = nodes->nodeTab;
+		ci = xkl_config_item_new();
+		for (j = nodes->nodeNr; --j >= 0;) {
+
+			if (xkl_read_config_item(config, di, *pnode, ci)) {
+				if (g_slist_find_custom
+				    (processed_ids, ci->name,
+				     (GCompareFunc) g_ascii_strcasecmp) ==
+				    NULL) {
+					gboolean allow_multisel = TRUE;
+					xmlChar *sallow_multisel =
+					    xmlGetProp(*pnode,
+						       (unsigned char *)
+						       XCI_PROP_ALLOW_MULTIPLE_SELECTION);
+					if (sallow_multisel != NULL) {
+						allow_multisel =
+						    !g_ascii_strcasecmp
+						    ("true", (char *)
+						     sallow_multisel);
+						xmlFree(sallow_multisel);
+						g_object_set_data(G_OBJECT
+								  (ci),
+								  XCI_PROP_ALLOW_MULTIPLE_SELECTION,
+								  GINT_TO_POINTER
+								  (allow_multisel));
+					}
+
+					func(config, ci, data);
+					processed_ids =
+					    g_slist_append(processed_ids,
+							   g_strdup
+							   (ci->name));
 				}
-
-				func(config, ci, data);
 			}
 
 			pnode++;
@@ -620,6 +756,8 @@ xkl_config_registry_foreach_option_group(XklConfigRegistry *
 		g_object_unref(G_OBJECT(ci));
 		xmlXPathFreeObject(xpath_obj);
 	}
+	g_slist_foreach(processed_ids, (GFunc) g_free, NULL);
+	g_slist_free(processed_ids);
 }
 
 void
@@ -721,14 +859,16 @@ xkl_config_rec_activate(const XklConfigRec * data, XklEngine * engine)
 }
 
 gboolean
-xkl_config_registry_load(XklConfigRegistry * config)
+xkl_config_registry_load(XklConfigRegistry * config,
+			 gboolean if_extras_needed)
 {
 	XklEngine *engine;
 
 	xkl_config_registry_free(config);
 	engine = xkl_config_registry_get_engine(config);
 	xkl_engine_ensure_vtable_inited(engine);
-	return xkl_engine_vcall(engine, load_config_registry) (config);
+	return xkl_engine_vcall(engine, load_config_registry) (config,
+							       if_extras_needed);
 }
 
 gboolean
@@ -861,6 +1001,8 @@ xkl_config_registry_finalize(GObject * obj)
 extern void
 xkl_config_registry_class_term(XklConfigRegistryClass * klass)
 {
+	gint i;
+
 	if (models_xpath != NULL) {
 		xmlXPathFreeCompExpr(models_xpath);
 		models_xpath = NULL;
@@ -873,6 +1015,24 @@ xkl_config_registry_class_term(XklConfigRegistryClass * klass)
 		xmlXPathFreeCompExpr(option_groups_xpath);
 		option_groups_xpath = NULL;
 	}
+	if (xml_encode_regexen != NULL) {
+		for (i =
+		     sizeof(xml_encode_regexen_str) /
+		     sizeof(xml_encode_regexen_str[0]); --i >= 0;) {
+			g_regex_unref(xml_encode_regexen[i]);
+		}
+		g_free(xml_encode_regexen);
+		xml_encode_regexen = NULL;
+	}
+	if (xml_decode_regexen != NULL) {
+		for (i =
+		     sizeof(xml_decode_regexen_str) /
+		     sizeof(xml_decode_regexen_str[0]); --i >= 0;) {
+			g_regex_unref(xml_decode_regexen[i]);
+		}
+		g_free(xml_decode_regexen);
+		xml_decode_regexen = NULL;
+	}
 }
 
 static void
@@ -880,6 +1040,7 @@ xkl_config_registry_class_init(XklConfigRegistryClass * klass)
 {
 	GObjectClass *object_class;
 	GParamSpec *engine_param_spec;
+	gint i;
 
 	object_class = (GObjectClass *) klass;
 	parent_class = g_type_class_peek_parent(object_class);
@@ -909,4 +1070,21 @@ xkl_config_registry_class_init(XklConfigRegistryClass * klass)
 					XKBCR_LAYOUT_PATH);
 	option_groups_xpath = xmlXPathCompile((unsigned char *)
 					      XKBCR_GROUP_PATH);
+
+	xml_encode_regexen =
+	    g_new0(GRegex *,
+		   sizeof(xml_encode_regexen_str) /
+		   sizeof(xml_encode_regexen_str[0]));
+	xml_decode_regexen =
+	    g_new0(GRegex *,
+		   sizeof(xml_decode_regexen_str) /
+		   sizeof(xml_decode_regexen_str[0]));
+	for (i =
+	     sizeof(xml_encode_regexen_str) /
+	     sizeof(xml_encode_regexen_str[0]); --i >= 0;) {
+		xml_encode_regexen[i] =
+		    g_regex_new(xml_encode_regexen_str[i], 0, 0, NULL);
+		xml_decode_regexen[i] =
+		    g_regex_new(xml_decode_regexen_str[i], 0, 0, NULL);
+	}
 }
