@@ -47,7 +47,8 @@ typedef struct {
 } CodeBuildStruct;
 
 static const char *countryLookupNames[] = { "alpha_2_code", NULL };
-static const char *languageLookupNames[] = { "iso_639_2B_code", "iso_639_2T_code", NULL };
+static const char *languageLookupNames[] =
+    { "iso_639_2B_code", "iso_639_2T_code", NULL };
 
 static LookupParams countryLookup = { "iso_3166", countryLookupNames };
 static LookupParams languageLookup = { "iso_639", languageLookupNames };
@@ -96,12 +97,10 @@ iso_codes_parse_start_tag(GMarkupParseContext * ctx,
 		while (*attr) {
 			if (g_str_equal(*attr_names, *attr)) {
 				if (**attr_values) {
-					g_hash_table_insert(cbs->
-							    code_names,
-							    g_strdup
-							    (*attr_values),
-							    g_strdup
-							    (name));
+					g_hash_table_insert
+					    (cbs->code_names,
+					     g_strdup(*attr_values),
+					     g_strdup(name));
 				}
 			}
 			attr++;
@@ -212,6 +211,7 @@ xkl_config_registry_foreach_iso_code(XklConfigRegistry * config,
 	const gchar **xpath_expr;
 	gpointer key, value;
 	XklConfigItem *ci;
+	gint di;
 
 	if (!xkl_config_registry_is_initialized(config))
 		return;
@@ -219,38 +219,50 @@ xkl_config_registry_foreach_iso_code(XklConfigRegistry * config,
 	code_pairs = g_hash_table_new(g_str_hash, g_str_equal);
 
 	for (xpath_expr = xpath_exprs; *xpath_expr; xpath_expr++) {
-		xpath_obj =
-		    xmlXPathEval((unsigned char *) *xpath_expr,
-				 xkl_config_registry_priv(config,
-							  xpath_context));
-		if (xpath_obj != NULL) {
+		for (di = 0; di < XKL_NUMBER_OF_REGISTRY_DOCS; di++) {
 			gint ni;
-			xmlNodeSetPtr nodes = xpath_obj->nodesetval;
-			if (nodes != NULL) {
-				xmlNodePtr *node = nodes->nodeTab;
-				for (ni = nodes->nodeNr; --ni >= 0;) {
-					gchar *iso_code =
-					    (gchar *) (*node)->
-					    children->content;
-					const gchar *description;
-					iso_code =
-					    to_upper ?
-					    g_ascii_strup(iso_code,
-							  -1) :
-					    g_strdup(iso_code);
-					description = dgf(iso_code);
-/* If there is a mapping to some ISO description - consider it as ISO code (well, it is just an assumption) */
-					if (description)
-						g_hash_table_insert
-						    (code_pairs,
-						     g_strdup
-						     (iso_code),
-						     g_strdup
-						     (description));
-					g_free(iso_code);
-					node++;
-				}
+			xmlNodePtr *node;
+			xmlNodeSetPtr nodes;
+
+			xmlXPathContextPtr xmlctxt =
+			    xkl_config_registry_priv(config,
+						     xpath_contexts[di]);
+			if (xmlctxt == NULL)
+				continue;
+
+			xpath_obj =
+			    xmlXPathEval((unsigned char *) *xpath_expr,
+					 xmlctxt);
+			if (xpath_obj == NULL)
+				continue;
+
+			nodes = xpath_obj->nodesetval;
+			if (nodes == NULL) {
+				xmlXPathFreeObject(xpath_obj);
+				continue;
 			}
+
+			node = nodes->nodeTab;
+			for (ni = nodes->nodeNr; --ni >= 0;) {
+				gchar *iso_code =
+				    (gchar *) (*node)->children->content;
+				const gchar *description;
+				iso_code =
+				    to_upper ?
+				    g_ascii_strup(iso_code,
+						  -1) : g_strdup(iso_code);
+				description = dgf(iso_code);
+/* If there is a mapping to some ISO description - consider it as ISO code (well, it is just an assumption) */
+				if (description)
+					g_hash_table_insert
+					    (code_pairs,
+					     g_strdup
+					     (iso_code),
+					     g_strdup(description));
+				g_free(iso_code);
+				node++;
+			}
+
 			xmlXPathFreeObject(xpath_obj);
 		}
 	}
@@ -258,10 +270,9 @@ xkl_config_registry_foreach_iso_code(XklConfigRegistry * config,
 	g_hash_table_iter_init(&iter, code_pairs);
 	ci = xkl_config_item_new();
 	while (g_hash_table_iter_next(&iter, &key, &value)) {
-		g_snprintf(ci->name, sizeof(ci->name),
-			   (const gchar *) key);
-		g_snprintf(ci->description, sizeof(ci->description),
-			   (const gchar *) value);
+		g_strlcpy(ci->name, (const gchar *) key, sizeof(ci->name));
+		g_strlcpy(ci->description, (const gchar *) value,
+			  sizeof(ci->description));
 		func(config, ci, data);
 	}
 	g_object_unref(G_OBJECT(ci));
@@ -318,6 +329,7 @@ xkl_config_registry_foreach_iso_variant(XklConfigRegistry *
 					should_code_be_lowered2[])
 {
 	xmlXPathObjectPtr xpath_obj;
+	xmlNodeSetPtr nodes;
 	const gchar **xpath_expr;
 	const gboolean *is_low_id = should_code_be_lowered1;
 	gchar *low_iso_code;
@@ -331,21 +343,44 @@ xkl_config_registry_foreach_iso_variant(XklConfigRegistry *
 	     xpath_expr++, is_low_id++) {
 		const gchar *aic = *is_low_id ? low_iso_code : iso_code;
 		gchar *xpe = g_strdup_printf(*xpath_expr, aic);
-		xpath_obj =
-		    xmlXPathEval((unsigned char *) xpe,
-				 xkl_config_registry_priv(config,
-							  xpath_context));
-		if (xpath_obj != NULL) {
-			xmlNodeSetPtr nodes = xpath_obj->nodesetval;
+		gint di;
+		GSList *processed_ids = NULL;
+
+		for (di = 0; di < XKL_NUMBER_OF_REGISTRY_DOCS; di++) {
+			xmlXPathContextPtr xmlctxt =
+			    xkl_config_registry_priv(config,
+						     xpath_contexts[di]);
+			if (xmlctxt == NULL)
+				continue;
+
+			xpath_obj =
+			    xmlXPathEval((unsigned char *) xpe, xmlctxt);
+			if (xpath_obj == NULL)
+				continue;
+
+			nodes = xpath_obj->nodesetval;
 			if (nodes != NULL) {
 				gint ni;
 				xmlNodePtr *node = nodes->nodeTab;
 				XklConfigItem *ci = xkl_config_item_new();
 				for (ni = nodes->nodeNr; --ni >= 0;) {
 					if (xkl_read_config_item
-					    (config, *node, ci))
-						func(config, ci, NULL,
-						     data);
+					    (config, di, *node, ci)) {
+						if (g_slist_find_custom
+						    (processed_ids,
+						     ci->name,
+						     (GCompareFunc)
+						     g_ascii_strcasecmp) ==
+						    NULL) {
+							func(config, ci,
+							     NULL, data);
+							processed_ids =
+							    g_slist_append
+							    (processed_ids,
+							     g_strdup
+							     (ci->name));
+						}
+					}
 					node++;
 				}
 				g_object_unref(G_OBJECT(ci));
@@ -360,12 +395,20 @@ xkl_config_registry_foreach_iso_variant(XklConfigRegistry *
 	     xpath_expr++, is_low_id++) {
 		const gchar *aic = *is_low_id ? low_iso_code : iso_code;
 		gchar *xpe = g_strdup_printf(*xpath_expr, aic);
-		xpath_obj =
-		    xmlXPathEval((unsigned char *) xpe,
-				 xkl_config_registry_priv(config,
-							  xpath_context));
-		if (xpath_obj != NULL) {
-			xmlNodeSetPtr nodes = xpath_obj->nodesetval;
+		gint di;
+		for (di = 0; di < XKL_NUMBER_OF_REGISTRY_DOCS; di++) {
+			xmlXPathContextPtr xmlctxt =
+			    xkl_config_registry_priv(config,
+						     xpath_contexts[di]);
+			if (xmlctxt == NULL)
+				continue;
+
+			xpath_obj =
+			    xmlXPathEval((unsigned char *) xpe, xmlctxt);
+			if (xpath_obj == NULL)
+				continue;
+
+			nodes = xpath_obj->nodesetval;
 			if (nodes != NULL) {
 				gint ni;
 				xmlNodePtr *node = nodes->nodeTab;
@@ -373,9 +416,9 @@ xkl_config_registry_foreach_iso_variant(XklConfigRegistry *
 				XklConfigItem *pci = xkl_config_item_new();
 				for (ni = nodes->nodeNr; --ni >= 0;) {
 					if (xkl_read_config_item
-					    (config, *node, ci) &&
+					    (config, di, *node, ci) &&
 					    xkl_read_config_item
-					    (config,
+					    (config, di,
 					     (*node)->parent->parent, pci))
 						func(config, pci, ci,
 						     data);
