@@ -100,12 +100,14 @@ xkl_engine_process_focus_in_evt(XklEngine * engine,
 				XFocusChangeEvent * fev)
 {
 	Window win;
+	Window prev_toplevel_win =
+	    xkl_engine_priv(engine, curr_toplvl_win);
 	Window toplevel_win;
 	XklState selected_window_state;
 
 	if (!
-	    (xkl_engine_priv(engine, listener_type) &
-	     XKLL_MANAGE_WINDOW_STATES))
+	    (xkl_engine_is_listening_for
+	     (engine, XKLL_MANAGE_WINDOW_STATES)))
 		return;
 
 	win = fev->window;
@@ -135,36 +137,51 @@ xkl_engine_process_focus_in_evt(XklEngine * engine,
 
 	if (xkl_engine_get_toplevel_window_state
 	    (engine, toplevel_win, &selected_window_state)) {
-		if (xkl_engine_priv(engine, curr_toplvl_win) !=
-		    toplevel_win) {
-			gboolean old_win_transparent, new_win_transparent;
-			XklState tmp_state;
-
-			old_win_transparent =
-			    xkl_engine_is_toplevel_window_transparent
-			    (engine,
-			     xkl_engine_priv(engine, curr_toplvl_win));
-			if (old_win_transparent)
-				xkl_debug(150,
-					  "Leaving transparent window\n");
+		if (prev_toplevel_win != toplevel_win) {
+			gboolean new_win_transparent;
+			Window parent = (Window) NULL, root =
+			    (Window) NULL, *children = NULL;
+			guint nchildren = 0;
 
 			/*
-			 * Reload the current state from the current window. 
-			 * Do not do it for transparent window - we keep the state from 
-			 * the _previous_ window.
+			 * If previous focused window exists - handle transparency and state
+			 * (optional)
 			 */
-			if (!old_win_transparent &&
-			    xkl_engine_get_toplevel_window_state(engine,
-								 xkl_engine_priv
-								 (engine,
-								  curr_toplvl_win),
-								 &tmp_state))
-			{
-				xkl_engine_update_current_state(engine,
-								tmp_state.group,
-								tmp_state.indicators,
-								"Loading current (previous) state from the current (previous) window");
-			}
+			if (xkl_engine_query_tree
+			    (engine, prev_toplevel_win, &root, &parent,
+			     &children, &nchildren) == Success) {
+				XklState tmp_state;
+				gboolean old_win_transparent =
+				    xkl_engine_is_toplevel_window_transparent
+				    (engine, prev_toplevel_win);
+
+				if (children != NULL)
+					XFree(children);
+
+				if (old_win_transparent)
+					xkl_debug(150,
+						  "Leaving transparent window\n");
+				/*
+				 * Reload the current state from the current window. 
+				 * Do not do it for transparent window - we keep the state from 
+				 * the _previous_ window.
+				 */
+				if (!old_win_transparent
+				    &&
+				    xkl_engine_get_toplevel_window_state
+				    (engine, prev_toplevel_win,
+				     &tmp_state)) {
+					xkl_engine_update_current_state
+					    (engine, tmp_state.group,
+					     tmp_state.indicators,
+					     "Loading current (previous) state from the current (previous) window");
+				}
+			} else
+				xkl_debug(150,
+					  "Current (previous) window "
+					  WINID_FORMAT
+					  " does not exist any more, so transparency/state are not analyzed\n",
+					  prev_toplevel_win);
 
 			xkl_engine_priv(engine, curr_toplvl_win) =
 			    toplevel_win;
@@ -308,8 +325,8 @@ xkl_engine_process_focus_out_evt(XklEngine * engine,
 				 XFocusChangeEvent * fev)
 {
 	if (!
-	    (xkl_engine_priv(engine, listener_type) &
-	     XKLL_MANAGE_WINDOW_STATES))
+	    (xkl_engine_is_listening_for
+	     (engine, XKLL_MANAGE_WINDOW_STATES)))
 		return;
 
 	if (fev->mode != NotifyNormal) {
@@ -352,7 +369,6 @@ xkl_engine_process_focus_out_evt(XklEngine * engine,
 void
 xkl_engine_process_property_evt(XklEngine * engine, XPropertyEvent * pev)
 {
-	guint listener_type = xkl_engine_priv(engine, listener_type);
 	if (400 <= xkl_debug_level) {
 		char *atom_name =
 		    XGetAtomName(xkl_engine_get_display(engine),
@@ -371,7 +387,8 @@ xkl_engine_process_property_evt(XklEngine * engine, XPropertyEvent * pev)
 	}
 
 	if (pev->atom == xkl_engine_priv(engine, atoms)[WM_STATE]) {
-		if (listener_type & XKLL_MANAGE_WINDOW_STATES) {
+		if (xkl_engine_is_listening_for
+		    (engine, XKLL_MANAGE_WINDOW_STATES)) {
 			gboolean has_xkl_state =
 			    xkl_engine_get_state(engine, pev->window,
 						 NULL);
@@ -402,9 +419,12 @@ xkl_engine_process_property_evt(XklEngine * engine, XPropertyEvent * pev)
 		}		/* XKLL_MANAGE_WINDOW_STATES */
 	} else if (pev->atom == xkl_engine_priv(engine, base_config_atom)
 		   && pev->window == xkl_engine_priv(engine, root_window)) {
-		if (listener_type &
-		    (XKLL_MANAGE_WINDOW_STATES |
-		     XKLL_TRACK_KEYBOARD_STATE)) {
+		if (xkl_engine_is_listening_for
+		    (engine,
+		     XKLL_MANAGE_WINDOW_STATES) |
+		    xkl_engine_is_listening_for(engine,
+						XKLL_TRACK_KEYBOARD_STATE))
+		{
 			if (pev->state == PropertyNewValue) {
 				/* If root window got new *_NAMES_PROP_ATOM -
 				   it most probably means new keyboard config is loaded by somebody */
@@ -423,9 +443,8 @@ void
 xkl_engine_process_create_window_evt(XklEngine * engine,
 				     XCreateWindowEvent * cev)
 {
-	if (!
-	    (xkl_engine_priv(engine, listener_type) &
-	     XKLL_MANAGE_WINDOW_STATES))
+	if (!xkl_engine_is_listening_for
+	    (engine, XKLL_MANAGE_WINDOW_STATES))
 		return;
 
 	xkl_debug(200,
@@ -475,8 +494,9 @@ xkl_process_error(Display * dpy, XErrorEvent * evt)
 		xkl_engine_priv(engine, last_error_code) = evt->error_code;
 
 	switch (evt->error_code) {
-	case BadWindow:
 	case BadAccess:
+	case BadDrawable:
+	case BadWindow:
 		{
 			XGetErrorText(evt->display, evt->error_code, buf,
 				      sizeof(buf));
@@ -491,6 +511,22 @@ xkl_process_error(Display * dpy, XErrorEvent * evt)
 			break;
 		}
 	default:
+		if (engine != NULL
+		    && xkl_engine_priv(engine, process_x_error)) {
+			if (xkl_engine_priv(engine, process_x_error)
+			    (engine, evt)) {
+				xkl_debug(200,
+					  "X ERROR processed by the engine: %p, "
+					  WINID_FORMAT ", %d [%s], "
+					  "X11 request: %d, minor code: %d\n",
+					  dpy,
+					  (unsigned long) evt->resourceid,
+					  (int) evt->error_code, buf,
+					  (int) evt->request_code,
+					  (int) evt->minor_code);
+				break;
+			}
+		}
 		xkl_debug(200,
 			  "Unexpected by libxklavier X ERROR: %p, "
 			  WINID_FORMAT ", %d [%s], "
@@ -537,15 +573,14 @@ xkl_engine_process_state_modification(XklEngine * engine,
 	 */
 	if (!xkl_engine_find_toplevel_window
 	    (engine, focused, &focused_toplevel)
-	    && xkl_engine_priv(engine,
-			       listener_type) & XKLL_MANAGE_WINDOW_STATES)
+	    && xkl_engine_is_listening_for(engine,
+					   XKLL_MANAGE_WINDOW_STATES))
 		focused_toplevel = xkl_engine_priv(engine, curr_toplvl_win);	/* what else can I do */
 
 	xkl_debug(150, "Focused window: " WINID_FORMAT ", '%s'\n",
 		  focused_toplevel,
 		  xkl_get_debug_window_title(engine, focused_toplevel));
-	if (xkl_engine_priv(engine, listener_type) &
-	    XKLL_MANAGE_WINDOW_STATES) {
+	if (xkl_engine_is_listening_for(engine, XKLL_MANAGE_WINDOW_STATES)) {
 		xkl_debug(150, "CurClient: " WINID_FORMAT ", '%s'\n",
 			  xkl_engine_priv(engine, curr_toplvl_win),
 			  xkl_get_debug_window_title(engine,
@@ -563,8 +598,8 @@ xkl_engine_process_state_modification(XklEngine * engine,
 				xkl_engine_update_current_state(engine,
 								grp, inds,
 								"Updating the state from new focused window");
-				if (xkl_engine_priv(engine, listener_type)
-				    & XKLL_MANAGE_WINDOW_STATES)
+				if (xkl_engine_is_listening_for
+				    (engine, XKLL_MANAGE_WINDOW_STATES))
 					xkl_engine_add_toplevel_window
 					    (engine, focused_toplevel,
 					     (Window) NULL, FALSE,
@@ -618,8 +653,7 @@ xkl_engine_process_state_modification(XklEngine * engine,
 		xkl_engine_try_call_state_func(engine, change_type,
 					       &old_state);
 
-	if (xkl_engine_priv(engine, listener_type) &
-	    XKLL_MANAGE_WINDOW_STATES)
+	if (xkl_engine_is_listening_for(engine, XKLL_MANAGE_WINDOW_STATES))
 		xkl_engine_save_toplevel_window_state(engine,
 						      xkl_engine_priv
 						      (engine,
