@@ -96,7 +96,7 @@ xkl_xkb_resume_listen(XklEngine * engine)
 			      XkbNamesNotify, XKB_NAMES_EVT_DTL_MASK,
 			      XKB_NAMES_EVT_DTL_MASK);
 #ifdef HAVE_XINPUT
-	if (xkl_engine_priv(engine, features) | XKLF_DEVICE_DISCOVERY) {
+	if (xkl_engine_priv(engine, features) & XKLF_DEVICE_DISCOVERY) {
 		DevicePresence(display, xitype, xiclass);
 		XSelectExtensionEvent(display,
 				      xkl_engine_priv(engine, root_window),
@@ -353,8 +353,8 @@ xkl_xkb_get_server_state(XklEngine * engine, XklState * current_state_out)
 				 &current_state_out->indicators))
 		current_state_out->indicators &=
 		    xkl_engine_backend(engine, XklXkb,
-				       cached_desc)->
-		    indicators->phys_indicators;
+				       cached_desc)->indicators->
+		    phys_indicators;
 	else
 		current_state_out->indicators = 0;
 }
@@ -402,8 +402,8 @@ xkl_xkb_set_indicator(XklEngine * engine, gint indicator_num, gboolean set)
 						     xkl_engine_backend
 						     (engine, XklXkb,
 						      device_id),
-						     cached->
-						     names->indicators
+						     cached->names->
+						     indicators
 						     [indicator_num], set,
 						     False, NULL);
 			else {
@@ -541,7 +541,7 @@ xkl_xkb_init(XklEngine * engine)
 #ifdef LIBXKBFILE_PRESENT
 	gint opcode;
 	gboolean xkl_xkb_ext_present;
-	int xi_opc, xi_event_type, xi_error_code;
+	int xi_opc;
 
 	xkl_engine_priv(engine, backend_id) = "XKB";
 	xkl_engine_priv(engine, features) = XKLF_CAN_TOGGLE_INDICATORS |
@@ -564,6 +564,7 @@ xkl_xkb_init(XklEngine * engine)
 	xkl_engine_priv(engine, get_num_groups) = xkl_xkb_get_num_groups;
 	xkl_engine_priv(engine, lock_group) = xkl_xkb_lock_group;
 	xkl_engine_priv(engine, process_x_event) = xkl_xkb_process_x_event;
+	xkl_engine_priv(engine, process_x_error) = xkl_xkb_process_x_error;
 	xkl_engine_priv(engine, free_all_info) = xkl_xkb_free_all_info;
 	xkl_engine_priv(engine, if_cached_info_equals_actual) =
 	    xkl_xkb_if_cached_info_equals_actual;
@@ -590,12 +591,8 @@ xkl_xkb_init(XklEngine * engine)
 								    XklXkb,
 								    error_code),
 						NULL, NULL);
-	if (!xkl_xkb_ext_present) {
-		XSetErrorHandler((XErrorHandler)
-				 xkl_engine_priv(engine,
-						 default_error_handler));
+	if (!xkl_xkb_ext_present)
 		return -1;
-	}
 
 	xkl_debug(160,
 		  "xkbEvenType: %X, xkbError: %X, display: %p, root: "
@@ -619,14 +616,24 @@ xkl_xkb_init(XklEngine * engine)
 		xkl_engine_priv(engine, features) |=
 		    XKLF_MULTIPLE_LAYOUTS_SUPPORTED;
 
+#if HAVE_XINPUT
 	if (XQueryExtension
 	    (display, "XInputExtension", &xi_opc,
-	     &xi_event_type, &xi_error_code)) {
+	     &xkl_engine_backend(engine, XklXkb, xi_event_type),
+	     &xkl_engine_backend(engine, XklXkb, xi_error_code))) {
 		xkl_debug(150, "XInputExtension found (%d, %d, %d)\n",
-			  xi_opc, xi_event_type, xi_error_code);
+			  xi_opc,
+			  xkl_engine_backend(engine, XklXkb,
+					     xi_event_type),
+			  xkl_engine_backend(engine, XklXkb,
+					     xi_error_code));
 		xkl_engine_priv(engine, features) |= XKLF_DEVICE_DISCOVERY;
-	} else
+	} else {
 		xkl_debug(0, "XInputExtension not found\n");
+		xkl_engine_backend(engine, XklXkb, xi_event_type) = -1;
+		xkl_engine_backend(engine, XklXkb, xi_error_code) = -1;
+	}
+#endif
 	return 0;
 #else
 	xkl_debug(160,
